/* USAGE: ./benchmark -n MATRIX_SIZE
 *
 * Parameters:
 *  -n ###		matrix size n
 *  -t blas		use blas instead of your own dgemm
 *  -t my		use your own dgemm
 *  -a ##		parameter a for init
 *  -b ##		parameter b for init
 *  -i ##		iterations
 */
#include "util.h"

#define PEAK_SINGLE_CORE 4400
#define PEAK_ALL_CORES 35200
#define MAX_TIMES 5

#define INIT_A (-1)
#define INIT_B (1)



int main(int argc, char** argv) {
    int n;
    double res;
    double a = INIT_A;
    double b = INIT_B;
    int max_times = MAX_TIMES;
    matmul_type matmul = MY;
    int flag;
    int plot_output = 0;

    while ((flag = getopt(argc, argv, "n:a:b:t:p")) != -1) {
        switch (flag) {
            case 'h':
                printf("%s -n ####\n", argv[0]);
                exit(0);
                break;
            case 'n':
                n = atoi(optarg);
                break;
            case 'i':
                max_times = atoi(optarg);
                break;
            case 'a':
                a = atof(optarg);
                break;
            case 'b':
                b = atof(optarg);
                break;
            case 't':
                if (strcmp(optarg, "blas") == 0) {
                    matmul = BLAS;
                } else {
                    matmul = MY;
                }
                break;
            case 'p':
                plot_output = 1;
                break;
        }
    }

    /* papi  */
    int chk = PAPI_NULL;
    float rt1, rt2, pt1, pt2, mfl;
    long long flo1, flo2;
    float rt[MAX_TIMES], perf[MAX_TIMES];


    /* Initialize the library */
    chk = PAPI_library_init(PAPI_VER_CURRENT);
    if (chk != PAPI_VER_CURRENT) {
        printf("PAPI library init error!\n");
        exit(1);
    }

    /* Matrices */
    double* A = (double*) malloc(sizeof(double)*n*n);
    double* B = (double*) malloc(sizeof(double)*n*n);
    double* C = (double*) malloc(sizeof(double)*n*n);

    initmatrix(n,a,b,A,B);

    /* run every size several times */
    for (int i=0; i<max_times; i++) {
        PAPI_flops(&rt1, &pt1, &flo1, &perf[i]);
        if (matmul == BLAS) {
            cblas_dgemm(CblasColMajor, CblasNoTrans, CblasNoTrans, n, n, n, 1e0, A, n, B, n, 0e0, C, n);
        } else {
            my_dgemm(n,A,B,C);
        }
        PAPI_flops(&rt2, &pt2, &flo2, &perf[i]);
        rt[i] = rt2 - rt1;
    }

    res = residual(n, A, B, C);

    if (plot_output) {
        printf(
                "%d %f %f %f %f %d %lld %e\n",
                n,
                min(MAX_TIMES,rt),
                max(MAX_TIMES,perf)/PEAK_SINGLE_CORE,
                max(MAX_TIMES,perf)/PEAK_ALL_CORES,
                max(MAX_TIMES,perf),
                PEAK_SINGLE_CORE,
                flo2-flo1,
                res
              );

    } else {
        printf("n: %d; a: %f; b: %f\n", n, a, b);
        printf(
                "runtime: %.5f sec; performance: %.2f%% %.2f%% (all cores); mflop/s: %.2f/%d; flop: %lld; residual: %e\n",
                min(MAX_TIMES,rt),
                max(MAX_TIMES,perf)/PEAK_SINGLE_CORE,
                max(MAX_TIMES,perf)/PEAK_ALL_CORES,
                max(MAX_TIMES,perf),
                PEAK_SINGLE_CORE,
                flo2-flo1,
                res
              );
    }
    free(A);
    free(B);
    free(C);
    return 0;
}
